import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.StringTokenizer;
import java.util.Vector;



/**
 * DeTable.Config.java
 *
 * reads the runtime-arguments from the command-line or (if specifed via command-line) from a file.
 *
 * @author guenther sommer <gue@rettung.at>
 */


public class Config
{
  String inFileName ="";
  String htmlFileName ="";
  String xmlFileName ="";
  int headerRow = -1;
  boolean supression = true;

  private Vector ConfigList; // list with all arguemnts
  private String ConfigArg;  // the actual argument
  private int ArgPos = -1; // actual position of arguemnts in work

  /**
   * goes through the list of arguments and sets the state-properties of the object as specifed through the arguments
   *
   * @param Args  arguments
   * @exception FileNotFoundException
   * @exception IOException
   */

  public Config(String[] Args) throws FileNotFoundException, IOException
  {
    ConfigList = new Vector(); // List
    if ((Args.length==0) || (Args[0].equals("-?")) || (Args[0].equals("--help")) )
      showHelpText();

    initArgs(Args); // reads arguemnts from command-line plus file arguements, if specifed in command-line
    while (!nextArg()) // go through all arguemnts
    {
      // configuration-file

      if ((ConfigArg.equals("-c")) || (ConfigArg.equals("--config")))
      {
        nextArg(); // get the word after the agrument
        fileArgs(ConfigArg); // ok, read the file and add to the argumenbt list
      }

      if ((ConfigArg.equals("-i")) || (ConfigArg.equals("--input-file")))
      {
        nextArg();
        inFileName = ConfigArg;
      }

      if ((ConfigArg.equals("-h")) || (ConfigArg.equals("--html-file")))
      {
        nextArg();
        htmlFileName = ConfigArg;
      }

      if ((ConfigArg.equals("-x")) || (ConfigArg.equals("--xml-file")))
      {
        nextArg();
        xmlFileName = ConfigArg;

      }

      if ((ConfigArg.equals("-r")) || (ConfigArg.equals("--header-row")))
      {
        nextArg();
        headerRow = Integer.parseInt(ConfigArg)+1;
      }

      if ((ConfigArg.equals("-ns")) || (ConfigArg.equals("--no-supression")))
      {
          supression=false;
      }
    }
  }


  /**
   * adds the args into the internal Args-List.
   *
   * @param args  the arguments
   */

  private void initArgs(String Args[])
  {
    for (int i = 0; i < Args.length; i++)
      ConfigList.addElement(Args[i]);
  }


  /**
   * the next Argument is stored in ConfigArg
   */

  private boolean nextArg()
  {
    ArgPos++;
    if (ArgPos >= (ConfigList.size()))
      return (true);
    ConfigArg = (String) ConfigList.elementAt(ArgPos);
    return (false);
  }


  /**
   * read arguemnts from a config file and adds them to the config list.
   */

  private void fileArgs(String ConfigFileName) throws FileNotFoundException, IOException
  {
    BufferedReader config_file = new BufferedReader(new FileReader(ConfigFileName));

    String config_in = config_file.readLine();
    while (!(config_in == null))
    {
      StringTokenizer config_in_token = new StringTokenizer(config_in, " ");
      while (config_in_token.hasMoreTokens())
        ConfigList.addElement(config_in_token.nextToken());

      config_in = config_file.readLine();
    }
  }


  /**
   * prints out a helptext with arguments.
   */

  private void showHelpText()
  {
    System.out.println("    -h  --help");
    System.out.println("        shows helptext");
      System.out.println();
    System.out.println("    -c FILE   --config FILE");
    System.out.println("        filename with extra arguments, even the -c parameter can be used for another file");
    System.out.println();
    System.out.println("    -r ROW   --header-row ROW");
    System.out.println("        use this row as header row");
    System.out.println();
    System.out.println("    -s       --no-supression");
    System.out.println("        deactive supression by duplicated header cells");
    System.out.println();
    System.out.println("    -i FILE   --input-file FILE");
    System.out.println("        filename with the html file to parse");
    System.out.println();
    System.out.println("    -h FILE   --html-file FILE");
    System.out.println("        filename for html-output (of the splitted table)");
    System.out.println();
    System.out.println("    -x FILE   --xml-file FILE");
    System.out.println("        filename for the xml-otuput ");
    System.exit(0);
  }
}