//
// DeTable.java
//
// written by guenther sommer <gue@rettung.at>
// last update by guenther sommer <gue@rettung.at> 01 Sep 2004
//
// written for the Institute for Software Technology and Interactive Systems 
// at the Vienna University of Technology. 
// for information see: http://www.ifs.tuwien.ac.at/~kaiser
//

/**
 *
 * DeTable.java
 *
 * breaks up a given html table with connected cells to individual cells
 * output is in html including all tags before, in and after the table
 * and in xml (following the detable.dtd) , with all tags removed
 * for more information see <a href="http://www.ifs.tuwien.ac.at/~kaiser">ifs at tu vienna</a>
 *
 * @author guenther sommer <gue@rettung.at>
 *
 */

import java.io.*;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.Vector;



public class DeTable
{
    static Config ConfigArg;  // Configuration-data (like filenames, ...) for the programm

    static Vector listTokens; // list of all tokens (text or tag) after parsing
    static Vector listPreTableTokens = new Vector(); // list of all tokens (text or tag) before the table
    static Vector listPostTableTokens = new Vector(); // list of all tokens (text or tag) after the table

    static Vector listRows = new Vector(); // dynamic list, has a list of TableCells in it


    /**
     * breaks the file supplied by a parameter into tokens <br>
     * a token is a text (between tags) or a tag <br>
     * the delimetres are the < and the >
     *
     * @param inputFile       a buffererd reader for the input file
     * @exception  IOException
     * @return        returns a vetor list containig the tokens as strings
     */

    public static Vector HTMLTokenizer(BufferedReader inputFile) throws IOException
    {
        char charBuffer[] = new char[1]; // input buffer
        String token = ""; // token to build
        boolean inTag = false; // am i in a tag
        Vector listInputTokens = new Vector(); // tokens returned by the tokenizer
        int eof; // are we at end-of-file

        // check for end-of-file
        eof = inputFile.read(charBuffer);

        while (eof != -1)
        {
            char charIn = charBuffer[0];

            // start of a tag
            if (charIn == '<')
            {
                inTag = true;

                // if token string is not empty, add the token string
                if (!token.equals(""))
                {
                    listInputTokens.add(token);
                }

                // Start new Token string
                token = "" + charIn;
            }
            // end of a tag and already in a tag
            else if ((charIn == '>') && inTag)
            {
                // add token and clear token string
                token = token + charIn;
                listInputTokens.add(token);
                token = "";
                inTag = false;

            }
            // we are outside or inside a tag, add the character to the token until a < or > appears
            else
            {
                token = token + charIn;
            }
            // read and check for eof
            eof = inputFile.read(charBuffer);

            // if a CR or LF is there, throw it away and read next character
            while (((charBuffer[0] == 13) || (charBuffer[0] == 10)) && (eof != -1))
            {
                eof = inputFile.read(charBuffer);
            }
        }

        // return the resulting VECTOR list
        return listInputTokens;
    }

    /**
     * moves the tokens before and after the table into seperate lists<br>
     * this tokens are removed from listTokens and put into listPreTableTokens and listPostTableTokens
     */

    public static void PeelTableTokens()
    {
        boolean tableStarted = false;
        boolean tableEnded = false;

        for (Iterator i = listTokens.iterator(); i.hasNext();)
        {
            String token = (String) i.next();

            // table not started and not ended --> before table --> move to pretable (including the table tag, this is why it is here)
            if ((!tableStarted) && (!tableEnded))
            {
                listPreTableTokens.add(token);
                i.remove();
            }
            // token begins with <table --> table starts

            if (token.toUpperCase().indexOf("<TABLE") == 0)
            {
                tableStarted = true;
            }


            // token begins with </table --> table ends
            if (token.toUpperCase().indexOf("</TABLE") == 0)
            {
                tableEnded = true;
            }

            // table had started and has ended --> after table --> move to posttable (including the /table tag, this is why it is here)
            if ((tableStarted) && (tableEnded))
            {
                listPostTableTokens.add(token);
                i.remove();
            }

        }
    }


    /**
     * gets a TableRow from the listRows object <br>
     * if the row is not existing (posY greater than size of listRows), the row is created<br>
     * this is to create a dynamic array with different amount of cells per row
     *
     * @param posX    number of row
     * @return        returns a TableRow object containing the selected or a new row
     */

    public static TableRow getRow(int posY)
    {
        // create rows if necessary to have existing Y position
        while (posY >= listRows.size())
        {
            listRows.add(new TableRow());
        }

        return (TableRow) listRows.get(posY);
    }


    /**
     * gets a TableCell from the listRows and the containing listCells object <br>
     * if the row/cell is not existing (posY or posX greater than the size of the list) the row/cell is created<br>
     * this is to create a dynamic array with different amount of cells per row
     *
     * @param posX    number of row
     * @param posY    cell of specified row
     * @return        returns a TableCell object containing the selected or a new cell
     */

    public static TableCell getCell(int posX, int posY)
    {
        TableRow row = getRow(posY);

        // create cells if necessary to have existing X position
        while (posX >= row.listCells.size())
        {
            row.listCells.add(new TableCell());
        }

        return (TableCell) row.listCells.get(posX);
    }


    /**
     * gets a TableCell from the listRows and the containing listCells object <br>
     * if the row/cell is not existing (posY or posX greater than the size of the list) the row/cell is created<br>
     * this is to create a dynamic array with different amount of cells per row
     *
     * @param posX    number of row
     * @param posY    cell of specified row
     * @return        returns the position of the next free cell
     */

    public static int getNextFreeCell(int posX, int posY)
    {
        TableCell cell = getCell(posX, posY);
        while (!cell.isEmpty)
        {
            posX++;
            cell = getCell(posX, posY);
        }
        return posX;
    }


    /**
     * this is the main parser, parses the tags in the listTokens to create a talbe-object with rows and cells<br>
     */

    public static void parseTable()
    {
        int rowPosition = 0; // actual x position in the parsed table
        int columnPosition = 0; // actual x position in the parsed table

        int SizeX = 1;  // size of cell in X (ROWSPAN)
        int SizeY = 1;  // size of table in Y (COLSPAN)
        boolean isHeader = false; // TD or TH as tag used
        String attributes = ""; // attributes supplied in the TD or TH tag
        String cellText = ""; // text between TD and /TD or TH and /TH
        String cellTextNoTag = ""; // text between TD and /TD or TH and /TH, but tags like b, i, ... removed
        boolean inCell = false; // are we inside a cell
        boolean inRow = false; // are we inside a row
        boolean parsed = false;

        // create first row
        listRows.add(new TableRow());

        // go through the hole token list
        for (int i = 0; i < listTokens.size(); i++)
        {
            // get token
            String token = (String) listTokens.get(i);
            parsed=false;

            // another table tag found
            if (token.toUpperCase().indexOf("<TABLE") == 0)
            {
                System.out.println("another TABLE tag found, can not handle nested tables, exiting!");
                System.exit(3); // error code for table problems
            }

            // end of cell tag found /TD or /TH or a new TD or TH while in cell
            if (((token.toUpperCase().indexOf("</TD") == 0) || (token.toUpperCase().indexOf("</TH") == 0))
                || (inCell && ((token.toUpperCase().indexOf("<TD") == 0) || (token.toUpperCase().indexOf("<TH") == 0)))
                || (inCell && (token.toUpperCase().indexOf("</TR") == 0))
                || (inCell && (token.toUpperCase().indexOf("<TR") == 0)))

            {
                // if cell is bigger then 1x1 (ROWSPAN or COLSPAN), set sizeX x sizeY cells
                for (int x = 0; x < SizeX; x++)
                {
                    for (int y = 0; y < SizeY; y++)
                    {
                        // set the cell ojbect
                        TableCell cell = getCell(columnPosition + x, rowPosition + y);
                        cell.text = cellText;
                        cell.textNoTags = cellTextNoTag;
                        cell.attributes = attributes;
                        cell.isHeader = isHeader;
                        cell.isEmpty = false;

                        // store if duplicated (COLSPAN)
                        if (x>0)
                        {
                            cell.duplicated=true;
                        }
                    }
                }
                // set the new position
                columnPosition = columnPosition + SizeX;
                inCell=false;
                parsed=true;
            }

            // a new cell (TD, TH) is found
            if ((token.toUpperCase().indexOf("<TD") == 0) || (token.toUpperCase().indexOf("<TH") == 0))
            {
                // initialize variables
                SizeX = 1;
                SizeY = 1;
                attributes = "";
                cellText = "";
                cellTextNoTag = "";

                // we ar in a cell, necessary, if a TD or TH without a / arrives
                inCell=true;

                // get the next free position in row for the cekk
                columnPosition = getNextFreeCell(columnPosition, rowPosition);

                // is it a header cell or a normal one
                if (token.toUpperCase().indexOf("<TH") == 0)
                {
                    isHeader = true;
                }
                else
                {
                    isHeader = false;
                }

                // remove starting tag and closing bracket
                StringTokenizer attribTokens = new StringTokenizer(token.substring(3, token.length() - 1));

                // search tokens in tag for rowspan and colspan parameters, parse and remove them, all others are added again
                while (attribTokens.hasMoreTokens())
                {
                    String tempToken = attribTokens.nextToken();

                    if (tempToken.toUpperCase().indexOf("COLSPAN=") == 0)
                    {
                        // remove COLSPAN= (8 chars) and convert to integer
                        SizeX = Integer.parseInt(tempToken.substring(8));
                    }
                    else if (tempToken.toUpperCase().indexOf("ROWSPAN=") == 0)
                    {
                        // remove ROWSPAN= (8 chars) and convert to integer
                        SizeY = Integer.parseInt(tempToken.substring(8));
                    }
                    else
                    {
                        // no colspan or rowspan attribute, so add it again
                        attributes = attributes + tempToken + " ";
                    }
                }

                parsed=true;
            }

            // a row ends
            if ((token.toUpperCase().indexOf("</TR") == 0) ||
                (inRow && (token.toUpperCase().indexOf("<TR") == 0)))
            {
                // update row, and position
                rowPosition++;
                columnPosition = 0;
                inRow=false;
                inCell=false;
                parsed=true;
            }

            // a new row begins
            if (token.toUpperCase().indexOf("<TR") == 0)
            {
                // update row, set attributes
                TableRow actualRow = getRow(rowPosition);
                actualRow.attributes = token.substring(3, token.length() - 1);
                inRow=true;
                parsed=true;
            }


            // no table tags, so add it to the cell text
            if (!parsed)
            {
                cellText = cellText + token;

                // don't add if a tag, but no table tag
                if (token.indexOf("<") != 0)
                {
                    cellTextNoTag= cellTextNoTag + token;
                }
            }

        }
    }


    /**
     * this analyzes the table and finds out if there is a header, and what are duplicated header fields and then mark the cells for supression
     */

    public static void analyzeTable()
    {
        if (ConfigArg.headerRow>listRows.size())
        {
            System.out.println("given header row is greater than the actual size of the table");
            System.exit(3); // error code for nested table
        }

        // find header row
        int headerRowPos=-1;
        if (ConfigArg.headerRow==-1)
        {
            for (int i=0;i<listRows.size();i++)
            {
                TableRow actualRow=getRow(i);

                for (int j=0;j<actualRow.listCells.size();j++)
                {
                    TableCell actualCell=getCell(j,i);
                    if (actualCell.isHeader)
                    {
                        // it's ok if no computed row found or the same
                        if ((headerRowPos==-1) || (headerRowPos==i))
                        {
                            // first row with headers
                            headerRowPos=i;
                        }
                        else
                        {
                            System.out.println("two or more rows with <TH> tags, try --no-supression");
                            System.exit(3); // error code for table problems
                        }
                    }
                }
            }
        }
        else
        {
            headerRowPos=ConfigArg.headerRow;
        }

        // check header row
        if (headerRowPos==-1)
        {
            System.out.println("no header row found, define one or try try --no-supression");
            System.exit(3); // error code for table problems
        }

        // supress header duplicated cells in table
        for (int i=0;i<listRows.size();i++)
        {
            TableRow actualRow=getRow(i);

            for (int j=0;j<actualRow.listCells.size();j++)
            {
                TableCell actualCell=getCell(j,i);
                // is column duplicated in header
                if (getCell(j,headerRowPos).duplicated)
                {
                    // is cell duplicated in actual row
                    if (actualCell.duplicated)
                    {
                        // supress
                        actualCell.supress=true;
                    }
                }
            }
        }
    }


    /**
     * writes the html output<br>
     * writes preTable tokens, the table created and the post table tokens<br>
     * tries to make the output html file very similiar to the input file
     *
     * @param outFile    PrintWriter as html output file
     */

    public static void writeTable(PrintWriter outFile)
    {
        // print out html before table
        for (int i = 0; i < listPreTableTokens.size(); i++)
        {
            outFile.print((String) listPreTableTokens.get(i));
        }

        // print out row by row
        for (int i = 0; i < listRows.size(); i++)
        {
            TableRow actualRow = getRow(i);

            // attributes used?
            if (actualRow.attributes.equals(""))
            {
                // no attributes
                outFile.print("<TR>");
            }
            else
            {
                // with attributes
                outFile.print("<TR " + actualRow.attributes + ">");
            }

            // print out cell by cell in a row
            for (int j = 0; j < actualRow.listCells.size(); j++)
            {
                TableCell actualCell = getCell(j, i);

                // is the cell a header cell
                if (actualCell.isHeader)
                {
                    // header cell
                    if (actualCell.attributes.equals(""))
                    {
                        // header cell without attributes
                        outFile.print("<TH>");
                    }
                    else
                    {
                        // header cell with attributes
                        outFile.print("<TH " + actualCell.attributes + ">");
                    }

                    // text in cell
                    outFile.print(actualCell.text);
                    // cell end
                    outFile.print("</TH>");
                }
                else if ((!actualCell.supress) || (!ConfigArg.supression))
                {
                    // no header cell
                    if (actualCell.attributes.equals(""))
                    {
                        // cell without attributes
                        outFile.print("<TD>");
                    }
                    else
                    {
                        // cell with attributes
                        outFile.print("<TD " + actualCell.attributes + ">");
                    }
                    outFile.print(actualCell.text);
                    outFile.print("</TD>");
                }
                else
                {
                    outFile.print("<TD></TD>");
                }

            }

            // end of row
            outFile.print("</TR>\r\n");
        }

        // print out html after table
        for (int i = 0; i < listPostTableTokens.size(); i++)
        {
            outFile.print((String) listPostTableTokens.get(i));
        }
    }

    /**
     * writes the xml output<br>
     * writes the the table size and the table created, but no html tags or pre/post text and tags <br>
     * tries to make the output xml as "textual" as possible
     *
     * @param outFile    PrintWriter as xml output file
     */

    public static void writeXml(PrintWriter outFile)
    {
        // print xml-header
        outFile.println("<?xml version=\"1.0\" encoding=\"UTF-8\" ?>");
        outFile.println("<!DOCTYPE table SYSTEM \"detable.dtd\">");

        outFile.println("<table>");

        // print maximum size of xml-table
        int MaxY = listRows.size();
        int MaxX = 0;
        for (int i = 0; i < listRows.size(); i++)
        {
            TableRow actualRow = getRow(i);
            // get biggest size of a row
            if (actualRow.listCells.size() > MaxX)
            {
                MaxX = actualRow.listCells.size();
            }
        }
        outFile.println("<size widthX=\"" + MaxX + "\" widthY=\"" + MaxY + "\"/>");

        // write xml table out, row by row
        for (int i = 0; i < listRows.size(); i++)
        {
            // write row start
            TableRow actualRow = getRow(i);
            outFile.println("<row>");

            /// print cell by cell
            for (int j = 0; j < actualRow.listCells.size(); j++)
            {
                TableCell actualCell = getCell(j, i);

                if (actualCell.isHeader)
                {
                    // header cell
                    outFile.print("<cell isHeader=\"true\">");
                }
                else
                {
                    // no header cell
                    outFile.print("<cell>");
                }

                if ((!actualCell.supress) || (!ConfigArg.supression))
                {
                    // don't print tags, which do formatting, ...
                    outFile.print(actualCell.textNoTags);
                }

                // cell end
                outFile.print("</cell>");
            }

            // row end
            outFile.println("</row>");
        }

        // file end
        outFile.println("</table>");
    }


    /**
     * the main routine<br>
     * reads the configuration, reads and parses the input file and writes (if configured) html and xml ouput<br>
     * @param args    Arguments from command line
     */

    public static void main(String args[])
    {
        // print copyleft information
        System.out.println("DeTable 1.1.3 - written by guenther sommer 2003-2004, contact gue-at-rettung.at");
        System.out.println("developed for the Institute for Software Technology and Interactive Systems");
	System.out.println("at the Vienna University of Technology");
        System.out.println("this app reads table in html file and splits up combined cells and puts it out as html or xml");
        System.out.println("licensed under the GPL v2, for further information see www.fsf.org");
        System.out.println();

        System.out.println("reading configuration ...");
        // try to read config file
        try
        {
            ConfigArg = new Config(args); // read the configuration-args or config-file
        }
        catch (FileNotFoundException e)
        {
            System.err.println("config file not found: " + e.getMessage());
            System.exit(2); // file error code
        }
        catch (IOException e)
        {
            System.err.println("error reading config file: " + e.getMessage());
            System.exit(2); // file error code
        }


        if (!ConfigArg.inFileName.equals(""))
        {
            System.out.println("reading input file ...");
            try
            {
                // open html file
                BufferedReader htmlInFile = new BufferedReader(new FileReader(ConfigArg.inFileName));

                // and read html file and create the tokens
                listTokens = HTMLTokenizer(htmlInFile);

                // close file
                htmlInFile.close();
            }
            catch (IOException e)
            {
                System.err.println("error reading input file: " + e.getMessage());
                System.exit(2); // file error code
            }
        }
        else
        {
            System.err.println("no input file provided");
            System.exit(1); // config error code
        }

        System.out.println("parsing table ...");

        // move tokens before and after table in two seperate lists
        PeelTableTokens();

        // and parse the table
        parseTable();

        // is supression enabled, then analyze table
        if (ConfigArg.supression)
        {
            analyzeTable();
        }

        // write html file if configured
        if (!ConfigArg.htmlFileName.equals(""))
        {
            System.out.println("writing html file ...");
            try
            {
                // open file for html output
                PrintWriter htmlOutFile = new PrintWriter(new BufferedWriter(new FileWriter(ConfigArg.htmlFileName)));

                // write the results into the file
                writeTable(htmlOutFile);

                // and close output file
                htmlOutFile.close();
            }
            catch (IOException e)
            {
                System.err.println("error writing html output file: " + e.getMessage());
                System.exit(2); // file error code
            }
        }

        // write xml file if configured
        if (!ConfigArg.xmlFileName.equals(""))
        {
            System.out.println("writing xml file ...");
            try
            {
                // open file for html output
                PrintWriter xmlOutFile = new PrintWriter(new BufferedWriter(new FileWriter(ConfigArg.xmlFileName)));

                // write the results into the file
                writeXml(xmlOutFile);

                // and close output file
                xmlOutFile.close();
            }
            catch (IOException e)
            {
                System.err.println("error writing xml output file: " + e.getMessage());
                System.exit(2); // file error code
            }
        }

        System.out.println("finished");
    }
}
